@@ -100,8 +100,8 @@ typedef unsigned long useconds_t;
  * Version of software...
  */
 
-#define CUPS_SVERSION "CUPS v2.4.14"
-#define CUPS_MINIMAL "CUPS/2.4.14"
+#define CUPS_SVERSION "CUPS v2.4.16"
+#define CUPS_MINIMAL "CUPS/2.4.16"
 
 
 /*
@@ -169,6 +169,13 @@ typedef unsigned long useconds_t;
 #define CUPS_DEFAULT_ERROR_POLICY "stop-printer"
 
 
+/*
+ * Default PeerCred value...
+ */
+
+#define CUPS_DEFAULT_PEER_CRED "on"
+
+
 /*
  * Default MaxCopies value...
  */
